/*
 *  Linux kernel adaptation layers for owl wifi library.
 *  Copyright (C) 2010 HD Wireless AB
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/types.h>

#include <wl_fw.h>
#include <wl_api.h>

#include "owl_net.h"
#include "owl_wext.h"
#include "owl_sync.h"

/*
 * The default behaviour for the driver is to load the firmware as soon
 * as the module is loaded and a card is inserterd. The device will be
 * enabled until the module is unloaded or the card removed.
 *
 * Hence, the link-level functions (scan, associate, etc) will be available
 * also when the device is considered 'down'. This behaviour is assumed by
 * some environments, e.g. android, while linux works either way.
 *
 * Therefore, the default is the allow the link-level stuff regardless of the
 * up/down state. This can be changed by defining OWL_CONFIG_IFUPDOWN.
 *
 */


/*
 * Invoked when the device should wake up from power save mode.
 * This function should send the proper sdio commands to the device.
 *
 */
static void wakeup_cb(void *ctx, uint8_t wakeup)
{
        struct owl_net *priv = ctx;
        priv->transport->wakeup(priv->transport, wakeup);
}


/*
 * Invoked when the a management message should be transmitted to the
 * WiFi device. This function should ensure that the message is passed through
 * to the device and should never fail.
 */
static void mgmt_tx_cb(void *ctx, const uint8_t *buf, uint16_t len)
{
        struct owl_net *priv = ctx;
        struct sk_buff *skb;

        skb = dev_alloc_skb(len);
        memcpy(skb_put(skb, len), buf, len);
        priv->transport->tx(priv->transport, skb);
}


/*
 * If OWL_CONFIG_IFUPDOWN is defined, this function will be called when
 * interface is about to be brought up, e.g upon 'ifconfig owl0 up'.
 * Won't be called unless interface is down.
 *
 * Otherwise this function will be called upon module load and card insertion.
 *
 * The device will be reset, firmware will be downloaded and wl_api will be
 * initialized each time this function is called.
 *
 */
static int owl_net_open(struct net_device *dev)
{
        wl_err_t wl_err;
        int err = 0;
        struct owl_net *priv = netdev_priv(dev);
        struct owl_transport *transport = priv->transport;

        dev_dbg(&dev->dev, "open\n");

        /* stuff used in owl_wext.c */
        priv->auth_mode = AUTH_MODE_OPEN_SYSTEM;
        priv->default_key = 0;
        memset(priv->key_info, 0, sizeof(priv->key_info));
        memset(priv->auth_param, 0, sizeof(priv->auth_param));
        priv->ps_traffic_timeout = 10; /* default */
        priv->ps_enabled = 0; /* default */
        
        if ((err = transport->fw_download(transport, fw_buf, fw_len))) {
                dev_err(&dev->dev, "fw download failed, err=%d\n", err);
                return err;
        }

        /* must be invoked before wl_init() */
        wl_register_mgmt_tx_cb(mgmt_tx_cb, priv);

        if ((err = owl_sync_wl_init(priv))) {
                owl_sync_wl_shutdown(priv);
                return err;
        }

        /* must be invoked before power save is enabled */
        wl_register_wakeup_cb(wakeup_cb, priv);
        
        if ((wl_err = wl_get_mac_addr(dev->dev_addr)) != WL_SUCCESS) {
                dev_err(&dev->dev, "get mac failed, err=%d\n", wl_err);
                owl_sync_wl_shutdown(priv);
                return -EIO;
        }

        /* carrier is only 'on' when associated. if this call is omitted
         * the stack will transmit an arp which we'll have to drop.
         */
        netif_carrier_off(dev);

        /* allow upper layers to call the device hard_start_xmit routine.
         * it won't be called until carrier is on though
         */
        netif_start_queue(dev);

        /* enable wl_poll() timer tick */
        mod_timer(&priv->wl_timer, jiffies + msecs_to_jiffies(100));
        
        return 0;
}


/*
 * If OWL_CONFIG_IFUPDOWN is defined, this function will be called when
 * interface is about to be brought down. Won't be called unless
 * interface is already up.
 *
 * Otherwise this function will be called upon module unload or card removal.
 *
 * All dynamic memory used by wl_api will be free'd.
 *
 */
static int owl_net_stop(struct net_device *dev)
{
        struct owl_net *priv = netdev_priv(dev);

        dev_dbg(&dev->dev, "ENTER %s\n", __func__);

        /* we don't want the wl_poll timer tick anymore */
        del_timer(&priv->wl_timer);

        /* Stop upper layers calling the device hard_start_xmit routine */
        netif_stop_queue(dev);
        owl_sync_wl_shutdown(priv);

        return 0;
}


/*
 * Invoked from owl_drv.c work queue when an incoming packet should be
 * processed. Will pass any data to netif.
 *
 */
void owl_net_rx(struct owl_transport *transport, struct sk_buff *skb)
{
        wl_err_t wl_err;
        char*  stripped_pkt;
        size_t stripped_pkt_len;
        uint16_t vlan;
        struct owl_net *priv = transport->net;
        struct net_device *dev = priv->dev;
        
        dev_dbg(&dev->dev, "owl_net_rx got pkt of size %d\n", skb->len);
        wl_err = owl_sync_wl_process_rx(priv,
                                        skb->data, skb->len, &stripped_pkt,
                                        &stripped_pkt_len, &vlan);

        /* As soon as any data is received, we'll try to enable tx again by
         * allowing the upper layers to call hard_start_xmit.
         *
         * The netif queue could have been stopped for two reasons, either the
         * tx queue on the card is full or the card was sleeping.
         *
         * If the card was sleeping, the fact that we just received something
         * means that the card is alive again so its ok to start the queue.
         *
         * If the tx queue was full, _and_ the received packet was a cfm,
         * indicating that a tx packet has been delivered by the card, there
         * is now room for a new packet.
         * 
         * However we don't know if the received packet was that kind of cfm,
         * so in some cases we might wake the queue even though the card is not
         * ready to accept another packet.
         *
         */
        netif_wake_queue(dev);
        
        switch (wl_err) {
        case WL_SUCCESS:
                dev_dbg(&dev->dev, "wl_process_rx returned WL_SUCCESS\n");
                skb_pull(skb, stripped_pkt - (char *) skb->data);
                skb_trim(skb, stripped_pkt_len);
                skb->protocol = eth_type_trans(skb, dev);
                priv->stats.rx_packets++;
                priv->stats.rx_bytes += stripped_pkt_len;
                netif_rx_ni(skb); /* invoked from workqueue */
                dev->last_rx = jiffies;
                return;

        case WL_ABSORBED:
                dev_dbg(&dev->dev, "wl_process_rx returned WL_ABSORBED\n");
                break;
                
        default:
                dev_warn(&dev->dev, "wl_process_rx failed, err=%d\n", wl_err);
#ifdef VERBOSE_DEBUG
                print_hex_dump_bytes("drop RX ", DUMP_PREFIX_NONE,
                                     skb->data, skb->len);
#endif
                break;
        };        

        dev_kfree_skb(skb);
}


/*
 * Called by upper layers when a packet should be transmitted. We can not sleep
 * in this function.
 *
 */
static int owl_net_tx(struct sk_buff *skb, struct net_device *dev)
{
        wl_err_t wl_err;
        struct owl_net *priv = netdev_priv(dev);
        struct sk_buff *skb_org = NULL;
           
        dev_dbg(&dev->dev, "ENTER %s, got pkt of size %d\n",
                __func__, skb->len);

        if (skb_headroom(skb) < WL_HEADER_SIZE) {
                struct sk_buff *skb2;
                dev_info(&dev->dev, "not enough headroom for wl hdr (%d<%d)\n",
                         skb_headroom(skb), WL_HEADER_SIZE);
                if ((skb2 = skb_realloc_headroom(skb, WL_HEADER_SIZE)) ==
                    NULL) {
                        dev_info(&dev->dev, "fail to allocate headroom\n");
                        goto drop_pkt;
                }
                /* don't free the original skb yet since we might have
                 * to retransmit it if wl_process_tx returns WL_RESOURCES
                 */
                skb_org = skb;
                skb = skb2;
        }
           
        skb_push(skb, WL_HEADER_SIZE);
        wl_err = owl_sync_wl_process_tx(
                priv, skb->data + WL_HEADER_SIZE, skb->len - WL_HEADER_SIZE,
                skb->len - WL_HEADER_SIZE, skb->data, 0);

        switch (wl_err) {
        case WL_SUCCESS:
                /* netdev watchdog compares dev->trans_start with jiffes,
                 * so make sure that trans_start is up to date.
                 * see net/sched/sch_generic.c.
                 */
                dev->trans_start = jiffies; 
                priv->transport->tx(priv->transport, skb);
                priv->stats.tx_packets++;
                priv->stats.tx_bytes += skb->len;
                if (skb_org) /* free original skb if cloned */
                        dev_kfree_skb(skb_org);

                /* skb will be free'd by transport layer */
                return NETDEV_TX_OK;

        case WL_AVAIL:
                dev_info(&dev->dev, "not connected\n");
#ifdef VERBOSE_DEBUG
                print_hex_dump_bytes("drop TX ", DUMP_PREFIX_NONE,
                                     skb->data, skb->len);
#endif
                goto drop_pkt;
                   
        case WL_RESOURCES:
                /* queue will be started on next rx */
                dev_dbg(&dev->dev, "tx busy, stopping queue\n");
                netif_stop_queue(dev);
                /* fall through */
                
        case WL_RETRY:
                dev_dbg(&dev->dev, "reschedule pkt\n");
                
                /* remove our extra headroom and free cloned skb,
                 * since original will be retransmitted
                 */
                if (skb_org) 
                        dev_kfree_skb(skb);
                else
                        skb_pull(skb, WL_HEADER_SIZE);
                
                return NETDEV_TX_BUSY;
                   
        default:
                dev_err(&dev->dev, "unexpected result %d\n", wl_err);
                goto drop_pkt;
        };
           

drop_pkt:
        dev_dbg(&dev->dev, "dropping pkt\n");
        dev_kfree_skb(skb);

        /* we had to clone the skb since there's was not enough headroom in the
         * original one. However the original skb could not be free'd until
         * we're done since it might have gotten retransmitted by upper layers.
         */
        if (skb_org)
                dev_kfree_skb(skb_org);
        
        priv->stats.tx_packets++;
        priv->stats.tx_dropped++;
        return NETDEV_TX_OK;
}


/*
 * Can be invoked at any time, even when the interface is down.
 *
 */
static struct net_device_stats *owl_net_get_stats(struct net_device *dev)
{
        struct owl_net *priv = netdev_priv(dev);
        return &priv->stats;
}


static void owl_net_tx_timeout(struct net_device *dev)
{
        dev_err(&dev->dev, "%s\n", __func__);
}


static const struct net_device_ops netdev_ops = {
#ifdef OWL_CONFIG_IFUPDOWN
        .ndo_open               = owl_net_open,
        .ndo_stop               = owl_net_stop,
#else
        .ndo_open               = NULL,
        .ndo_stop               = NULL,
#endif /* OWL_CONFIG_IFUPDOWN */
        .ndo_start_xmit         = owl_net_tx,
        .ndo_get_stats          = owl_net_get_stats,
        .ndo_set_multicast_list = NULL,
        .ndo_tx_timeout         = owl_net_tx_timeout,
        .ndo_set_mac_address    = NULL,
};


/*
 * Call owl_sync_wl_poll from process context since it might block.
 *
 */
static void owl_net_work(struct work_struct *work)
{
        struct owl_net *priv =
                container_of(work, struct owl_net, wl_work);
        struct net_device *dev = priv->dev;
        
        dev_vdbg(&dev->dev, "tick jiffies: %lu ms: %u\n", jiffies,
                 jiffies_to_msecs(jiffies));
        owl_sync_wl_poll(priv, jiffies_to_msecs(jiffies));
}


/*
 * Invoked periodically to schedule the timer tick required by wl_api
 *
 */
static void owl_net_timer_cb(unsigned long data)
{
        struct owl_net *priv = (struct owl_net *) data;
        schedule_work(&priv->wl_work);
        mod_timer(&priv->wl_timer, jiffies + msecs_to_jiffies(100));
}


/*
 * Called by sdio driver when the card is inserted and initialized.
 *
 * Will initialize the network interface. Note that this call will update the
 * net field in @transport.
 *
 */
int owl_net_register_drv(struct owl_transport *transport)
{
        int err;
        struct owl_net *priv;
        struct net_device *dev = NULL;
        
        BUG_ON(!transport);
        BUG_ON(!transport->tx);
        BUG_ON(!transport->fw_download);
        BUG_ON(transport->net);

        if ((dev = alloc_etherdev(sizeof(*priv))) == NULL)
                return -ENOMEM;

        priv = netdev_priv(dev);
        memset(priv, 0, sizeof(*priv));
        transport->net = priv;
        priv->dev = dev;
        priv->transport = transport;
        init_waitqueue_head(&priv->waitq);
	sema_init(&priv->wl_sema, 1);
        INIT_WORK(&priv->wl_work, owl_net_work);
        setup_timer(&priv->wl_timer, owl_net_timer_cb, (unsigned long) priv);
        
        if ((err = dev_alloc_name(dev, "owl%d")) < 0) {
                dev_err(&dev->dev, "dev_alloc_name(), err=%d\n", err); 
                goto err;
        }

        dev->netdev_ops = &netdev_ops;
        dev->watchdog_timeo = 8 * HZ;
        dev->wireless_handlers = &owl_wireless_handlers;
        dev->needed_headroom = WL_HEADER_SIZE;
        
        if ((err = register_netdev(dev))) {
                dev_err(&dev->dev, "register_netdev(), err=%d\n", err);
                goto err;
        }

#ifndef OWL_CONFIG_IFUPDOWN
        /* load firmare and configure device upon module load */
        if ((err = owl_net_open(dev)))
                goto err;
#endif /* OWL_CONFIG_IFUPDOWN */

        dev_info(&dev->dev, "ready\n");
        return 0;

err:
        free_netdev(dev);
        return err;
}


/*
 * Called by sdio driver when the card is removed or if the driver is unloaded.
 *
 */
void owl_net_unregister_drv(struct owl_transport *transport)
{
        struct owl_net *priv = transport->net;
        struct net_device *dev = NULL;
        
        BUG_ON(!priv);
        BUG_ON(!priv->dev);
        
        dev = priv->dev;
        dev_dbg(&dev->dev, "ENTER %s\n", __func__);

#ifndef OWL_CONFIG_IFUPDOWN
        (void) owl_net_stop(dev);
#endif /* OWL_CONFIG_IFUPDOWN */
         
        /* if OWL_CONFIG_IFUPDOWN is defined, owl_net_stop() is called if
         * necessary when rmmod, so we dont have to do wl_shutdown() here.
         */

        unregister_netdev(priv->dev);

        /* will also free priv */
        free_netdev(priv->dev);

        dev_info(&dev->dev, "shutdown\n");
}
