/*
 *  Linux kernel adaptation layers for owl wifi library.
 *  Copyright (C) 2010 HD Wireless AB
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/netdevice.h>
#include <linux/ctype.h>

#include "owl_wext_droid.h"
#include "owl_sync.h"

static int format_ssid(char *buf, struct wl_ssid_t *ssid);


/*
 * Android private ioctl handler.
 *
 * The Android patch to wpa_supplicant and wpa_cli will add some private
 * ioctls. The are all sent through SIOCSIWPRIV where the input buffer
 * contains a string describing the specific operation, and the reply
 * should be in text format, according to android_net_wifi_Wifi.cpp.
 *
 * The ioctl will be invoked from wpa_supplicant/driver_wext.c with the
 * following params:
 *
 *	struct iwreq iwr;
 * 	iwr.u.data.pointer = buf;
 *	iwr.u.data.length = buf_len;
 *
 * 'buf' can hold one of the following strings, it has a maximum length of
 * 248 (MAX_DRV_CMD_SIZE). Note that buf is an in and out parameter, it
 * should hold the response on success, maximum length is 'buf_len'.
 *
 *      RSSI           Return the current signal level in the format
 *                     "<ssid> rssi <value>". Note that RSSI-APPROX and RSSI
 *                     will both be sent as "RSSI" by driver_wext.c.
 *      COUNTRY US     Configure device to only use FCC channels (1-11). Note
 *                     that driver_wext.c will translate the SCAN-CHANNELS
 *                     command into a COUNTRY command.
 *      COUNTRY JP     Configure device to only use MKK1 channels (1-14)
 *      COUNTRY EU     Configure device to only use ETSI channels (1-13)
 *      SCAN-ACTIVE    Enable active scanning to be used for future scans.
 *      SCAN-PASSIVE   Enable passive scanning to be used for future scans.
 *      START          Start device. Unsure how this should be handled since
 *                     the device is started and stopped through ifup and
 *                     ifdown by the supplicant. See WifiStateTracker.java.
 *      STOP           Stop device, see above.
 *      RXFILTER-ADD 0     Unused
 *      RXFILTER-ADD 1     Unused
 *      RXFILTER-ADD 3     Unused
 *      RXFILTER-START     Unused. Start filtering out multicast packets.
 *                         The 4 start messages, will always be send in the
 *                         order above, see android_net_wifi_Wifi.cpp.
 *      RXFILTER-STOP      Unused. Stop filtering out multicast packets.
 *                         The 4 stop messages, will always be send in the
 *                         order below, see android_net_wifi_Wifi.cpp.
 *      RXFILTER-REMOVE 3  Unused
 *      RXFILTER-REMOVE 1  Unused
 *      RXFILTER-REMOVE 0  Unused
 *      LINKSPEED       Return the current link speed in the format
 *                      "LinkSpeed <mbps>", e.g. "LinkSpeed 54".
 *      MACADDR         Return our mac address in the format
 *                      "Macaddr = XX.XX.XX.XX.XX.XX".
 *      POWERMODE 0     Will be set by WifiStateTracker.java. '0' means
 *                      DRIVER_POWER_MODE_AUTO;
 *      POWERMODE 1     Will be set by WifiStateTracker.java. '1' means
 *                      DRIVER_POWER_MODE_ACTIVE;
 *      BTCOEXMODE 0    BLUETOOTH_COEXISTENCE_MODE_ENABLED.
 *      BTCOEXMODE 1    BLUETOOTH_COEXISTENCE_MODE_DISABLED.
 *      BTCOEXMODE 2    BLUETOOTH_COEXISTENCE_MODE_SENSE.
 *      BTCOEXSCAN-START
 *      BTCOEXSCAN-STOP
 *
 */
int owl_wext_set_droid(struct net_device *dev, struct iw_request_info *info,
                       union iwreq_data *wrqu, char *extra)
{
        char *ptr = wrqu->data.pointer;
        int len = wrqu->data.length;
        int err = 0;
        struct owl_net *priv = netdev_priv(dev);

        wrqu->data.length = 0; /* length of reply */
        
        if (ptr == NULL) {
                dev_err(&dev->dev, "in ptr is null, abort\n");
                return -EFAULT;
        }

        if (len == 0) {
                dev_err(&dev->dev, "length is 0, abort\n");
                return -EFAULT;
        }
        
        /* Note that the ioctl list (standard_ioctl) in net/wireless/wext.cs
         * declares the SIOCSIWPRIV to have a header type of
         * IW_HEADER_TYPE_NULL and not IW_HEADER_TYPE_POINT. Therefore,
         * we have to handle copy between user and kernel space (which is done
         * by wext.c for IW_HEADER_TYPE_POINT).
         */
	if (!(extra = kmalloc(len, GFP_KERNEL))) {
                dev_err(&dev->dev, "kmalloc() failed\n");
                return -ENOMEM;
        }
        
	if (copy_from_user(extra, ptr, len)) {
                dev_err(&dev->dev, "copy_from_user() failed\n");
                kfree(extra);
                return -EFAULT;
	}

        dev_dbg(&dev->dev, "ENTER %s cmd=\"%s\" len=%d\n", __func__, extra,
                len);

        if (strnicmp(extra, "RSSI", strlen("RSSI")) == 0) {

                char ssid[WL_SSID_MAX_LENGTH + 1];
                struct wl_network_t *net;
                
                if ((err = owl_sync_wl_get_current_network(priv, &net)) < 0) {
                        dev_err(&dev->dev,
                                "failed to get network, err=%d\n", err);
                        goto out;
                }

                if (net == NULL) {
                        dev_err(&dev->dev, "not connected when req rssi\n");
                        err = -EFAULT;
                        goto out;
                }
                
                format_ssid(ssid, &net->ssid);
                len = snprintf(extra, len, "%s rssi %d", ssid, net->rssi) + 1;
                kfree(net);
                
        } else if (strnicmp(extra, "LINKSPEED", strlen("LINKSPEED")) == 0) {

                wl_rate_t tx_rate;
                
                if (owl_sync_wl_get_rate(priv, &tx_rate, NULL) != WL_SUCCESS) {
                        dev_err(&dev->dev, "owl_sync_wl_get_rate() failed\n");
                        err = -EFAULT;
                        goto out;
                }
                        
                if (tx_rate == WL_RATE_INVALID) {
                        dev_err(&dev->dev, "invalid tx rate\n");
                        err = -EFAULT;
                        goto out;
                }

                len = snprintf(extra, len, "LinkSpeed %d", tx_rate / 2) + 1;
        
        } else if (strnicmp(extra, "MACADDR", strlen("MACADDR")) == 0) {

        	len = snprintf(extra, len,
                               "Macaddr = %02X:%02X:%02X:%02X:%02X:%02X\n",
                               dev->dev_addr[0], dev->dev_addr[1],
                               dev->dev_addr[2], dev->dev_addr[3],
                               dev->dev_addr[4], dev->dev_addr[5]) + 1;
        } else {

                dev_dbg(&dev->dev, "cmd \"%s\" has no effect\n", extra);
                goto out;
        }
        
        dev_dbg(&dev->dev, "replying with \"%s\"\n", extra);

        if (copy_to_user(ptr, extra, len)) {
                kfree(extra);
                return -EFAULT;
        }

        wrqu->data.length = len;
        
out:
        kfree(extra);
        return err;
}

static int format_ssid(char *buf, struct wl_ssid_t *ssid)
{
	int i, c;
	char *p = buf;

	if (ssid->len > 32) ssid->len = 32;

	for (i = 0; i < ssid->len; i++) {
		c = (int) ssid->ssid[i];
		if (c == '\\') {
			*p++ = '\\';
			*p++ = '\\';
		} else if (isprint((unsigned char) c)) {
			*p++ = (char) c;
		} else {
			p += sprintf(p, "\\x%02X", c);
		}
	}
	*p = '\0';

	return p - buf;
}
